﻿using System;
using System.Threading;
using System.Threading.Tasks;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using Magify.Model;
using UniRx;
using UnityEngine.Networking;

namespace Magify
{
    // ToDo: refactor ServerApi. Split it into separate classes
    internal partial class ServerApi
    {
        [NotNull]
        private readonly ServerRequestTools<GaladrielUser> _authorizeUserTools = new();
        [NotNull]
        private readonly ServerRequestTools<object> _saveStateTools = new();
        [NotNull]
        private readonly ServerRequestTools<string> _restoreStateTools = new();

        [NotNull]
        public IReactiveProperty<bool> SyncStateEnabled => _generalPrefs.SyncStateEnabled;

        #region Authorize

        /// <returns>
        /// User Id. Can be null
        /// </returns>
        public async UniTask<string> AuthorizeUser([NotNull] string clientId, CancellationToken cancellationToken)
        {
            _logger.Log($"{nameof(AuthorizeUser)} called");
            using var cancellationTokenSource = PrepareForRequest(_authorizeUserTools, cancellationToken);
            if (cancellationTokenSource == null)
            {
                return null;
            }
            var result = await Request(_authorizeUserTools, () => AuthorizeUserWithRepeat(clientId, cancellationToken), cancellationTokenSource);
            return result?.UserId;
        }

        public void CancelAllAuthorizeUserRequests()
        {
            _authorizeUserTools.CancelAllRequests();
            _authorizeUserTools.TrySetResult(null);
        }

        private async void AuthorizeUserWithRepeat([NotNull] string clientId, CancellationToken cancellationToken)
        {
            try
            {
                await RequestWithRepeat(requestAuthorizeUser, processResponse, _authorizeUserTools.Backoff, cancellationToken, _authorizeUserTools.RequestRepeatingDelayToken);
            }
            catch (OperationCanceledException)
            {
                _logger.LogWarning($"{nameof(AuthorizeUserWithRepeat)} was cancelled");
                _authorizeUserTools.TrySetResult(null);
            }

            async Task<WebResponseMessage> requestAuthorizeUser(CancellationToken cancellation)
            {
                _logger.Log($"{nameof(requestAuthorizeUser)} called");
                var requestMessage = CreateAuthorizeUserRequestMessage(clientId);
                _logger.Log($"Call {nameof(PerformRequestAsync)} with message and wait for result. Message: \n{requestMessage}");
                return await PerformRequestAsync(requestMessage, cancellation);
            }

            RepeatState processResponse(WebResponseMessage response)
            {
                _logger.Log($"{nameof(processResponse)} of {nameof(requestAuthorizeUser)} called");
                if (response.Result == UnityWebRequest.Result.Success)
                {
                    var authorizeUserResponse = JsonFacade.DeserializeObject<GaladrielResponse<GaladrielUser>>(response.Text)!;
                    _logger.Log($"User successfully authorized {authorizeUserResponse.Result}");
                    _authorizeUserTools.TrySetResult(authorizeUserResponse.Result);
                    return RepeatState.Finish;
                }

                return TryHandleDefaultTokenErrors(response, out var repeatState) ? repeatState : RepeatState.Wait;
            }
        }

        private WebRequestMessage CreateAuthorizeUserRequestMessage([NotNull] string clientId)
        {
            var authorizeUserPayload = new AuthorizeUserPayload
            {
                ClientId = clientId,
            };
            return new WebRequestMessage
            {
                AuthToken = default,
                Method = ApplicationStateApiMethods.Authorize(_platform.RuntimePlatform),
                Payload = authorizeUserPayload,
            };
        }

        #endregion

        #region SaveState

        public async UniTask<bool> SaveState([NotNull] string appState, [CanBeNull] int? weight, CancellationToken cancellationToken)
        {
            _logger.Log($"{nameof(SaveState)} called");
            using var cancellationTokenSource = PrepareForRequest(_saveStateTools, cancellationToken);
            if (cancellationTokenSource == null)
            {
                return false;
            }
            var result = await Request(_saveStateTools, () => SaveStateWithRepeat(appState, weight, cancellationToken), cancellationTokenSource);
            return result != null;
        }

        public void CancelAllSaveStateRequests()
        {
            _saveStateTools.CancelAllRequests();
            _saveStateTools.TrySetResult(null);
        }

        private async void SaveStateWithRepeat([NotNull] string appState, [CanBeNull] int? weight, CancellationToken cancellationToken)
        {
            try
            {
                await RequestWithRepeat(requestSaveState, processResponse, _saveStateTools.Backoff, cancellationToken, _saveStateTools.RequestRepeatingDelayToken);
            }
            catch (OperationCanceledException)
            {
                _logger.LogWarning($"{nameof(SaveStateWithRepeat)} was cancelled");
                _saveStateTools.TrySetResult(false);
            }

            async Task<WebResponseMessage> requestSaveState(CancellationToken cancellation)
            {
                _logger.Log($"{nameof(requestSaveState)} called");
                var requestMessage = CreateSaveStateRequestMessage(appState, weight);
                _logger.Log($"Call {nameof(PerformRequestAsync)} with message and wait for result. Message: \n{requestMessage}");
                return await PerformRequestAsync(requestMessage, cancellation);
            }

            RepeatState processResponse(WebResponseMessage response)
            {
                _logger.Log($"{nameof(processResponse)} of {nameof(requestSaveState)} called");
                if (response.Result == UnityWebRequest.Result.Success)
                {
                    _logger.Log($"State successfully saved");
                    _saveStateTools.TrySetResult(true);
                    return RepeatState.Finish;
                }

                return TryHandleDefaultTokenErrors(response, out var repeatState) ? repeatState : RepeatState.Wait;
            }
        }

        private WebRequestMessage CreateSaveStateRequestMessage([NotNull] string appState, [CanBeNull] int? weight)
        {
            var authorizeUserPayload = new SaveStatePayload()
            {
                Progress = appState,
                Weight = weight,
            };
            return new WebRequestMessage
            {
                AuthToken = default,
                Method = ApplicationStateApiMethods.SaveProgress(_platform.RuntimePlatform),
                Payload = authorizeUserPayload,
            };
        }

        #endregion

        #region RestoreState

        public async UniTask<string> RestoreState([CanBeNull] int? weight, CancellationToken cancellationToken)
        {
            _logger.Log($"{nameof(RestoreState)} called");
            using var cancellationTokenSource = PrepareForRequest(_restoreStateTools, cancellationToken);
            if (cancellationTokenSource == null)
            {
                return null;
            }
            return await Request(_restoreStateTools, () => RestoreStateWithRepeat(weight, cancellationToken), cancellationTokenSource);
        }

        public void CancelAllRestoreStateRequests()
        {
            _restoreStateTools.CancelAllRequests();
            _restoreStateTools.TrySetResult(null);
        }

        private async void RestoreStateWithRepeat([CanBeNull] int? weight, CancellationToken cancellationToken)
        {
            try
            {
                await RequestWithRepeat(requestRestoreState, processResponse, _restoreStateTools.Backoff, cancellationToken, _restoreStateTools.RequestRepeatingDelayToken);
            }
            catch (OperationCanceledException)
            {
                _logger.LogWarning($"{nameof(SaveStateWithRepeat)} was cancelled");
                _restoreStateTools.TrySetResult(null);
            }

            async Task<WebResponseMessage> requestRestoreState(CancellationToken cancellation)
            {
                _logger.Log($"{nameof(requestRestoreState)} called");
                var requestMessage = CreateRestoreStateRequestMessage(weight);
                _logger.Log($"Call {nameof(PerformRequestAsync)} with message and wait for result. Message: \n{requestMessage}");
                return await PerformRequestAsync(requestMessage, cancellation);
            }

            RepeatState processResponse(WebResponseMessage response)
            {
                _logger.Log($"{nameof(processResponse)} of {nameof(requestRestoreState)} called");
                if (response.Result == UnityWebRequest.Result.Success)
                {
                    _logger.Log($"State successfully restored. Result text={response.Text}");
                    GaladrielResponse<AppStateStateResponse> state;
                    try
                    {
                        state = JsonFacade.DeserializeObject<GaladrielResponse<AppStateStateResponse>>(response.Text);
                    }
                    catch (Exception e)
                    {
                        throw new MagifyFailedToParseRestoreProgressResponseException(response.Text, e);
                    }
                    _restoreStateTools.TrySetResult(state!.Result?.Progress);
                    return RepeatState.Finish;
                }

                return TryHandleDefaultTokenErrors(response, out var repeatState) ? repeatState : RepeatState.Wait;
            }
        }

        private WebRequestMessage CreateRestoreStateRequestMessage([CanBeNull] int? weight)
        {
            var authorizeUserPayload = new RestoreStatePayload()
            {
                Weight = weight,
            };
            return new WebRequestMessage
            {
                AuthToken = default,
                Method = ApplicationStateApiMethods.GetProgress(_platform.RuntimePlatform),
                Payload = authorizeUserPayload,
            };
        }

        #endregion

        #region Tools

        [CanBeNull]
        private CancellationTokenSource PrepareForRequest<T>([NotNull] ServerRequestTools<T> tools, CancellationToken cancellationToken)
        {
            if (_url.IsOffline)
            {
                _logger.Log("Offline mode is enabled. Request is skipped.");
                return null;
            }

            if (cancellationToken.IsCancellationRequested)
            {
                _logger.Log("Request skipped because it's cancelled");
                return null;
            }

            return CancellationTokenSource.CreateLinkedTokenSource(cancellationToken, _urlCancellationToken!.Token, tools.AllRequestsToken);
        }

        [ItemCanBeNull]
        private static async Task<T> Request<T>([NotNull] ServerRequestTools<T> tools, [NotNull] Action requestMethod, [NotNull] CancellationTokenSource cancellationTokenSource)
            where T: class
        {
            if (tools.TryCreateNewCompletionSource(out var completionTask))
            {
                _logger.Log($"Start requesting to server for {typeof(T).Name}");
                requestMethod();
            }
            else
            {
                completionTask = tools.CompletionTask;
                _logger.Log($"Request to server for {typeof(T).Name} is waiting for existing request result");
                // We have to reset delay, because new AuthorizeUser was triggered
                tools.ResetRequestsRepeatDelay();
            }

            _logger.Log("Wait till request result or cancel token");
            await Task.WhenAny(completionTask, cancellationTokenSource.Token.AsTask())!;
            if (cancellationTokenSource.IsCancellationRequested)
            {
                _logger.Log("Authorization token request was cancelled");
                return null;
            }

            return await completionTask!;
        }

        private bool TryHandleDefaultTokenErrors(WebResponseMessage response, out RepeatState repeatState)
        {
            if (!string.IsNullOrEmpty(response.Text) && response.ResponseCode == 403)
            {
                var errorResponse = JsonFacade.DeserializeObject<ErrorResponse>(response.Text);
                switch (errorResponse?.Error?.Code)
                {
                    case ErrorCode.Expired:
                        _logger.Log("Authorization Token expired, need reissue");
                        _authorizationToken!.IsExpired = true;
                        repeatState = RepeatState.Retry;
                        return true;

                    case ErrorCode.Revoked:
                        _logger.Log("Authorization Token revoked, trying to reset current token and request new one");
                        _authorizationToken!.AuthToken = null;
                        repeatState = RepeatState.Wait;
                        return true;

                    case ErrorCode.Banned:
                        HandleBannedResponse();
                        repeatState = RepeatState.Wait;
                        return true;
                }
            }
            repeatState = RepeatState.Wait;
            return false;
        }

        #endregion
    }
}