﻿using System;
using System.Collections.Generic;
using System.Threading;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using Magify.Model;
using UniRx;
using UnityEngine;
using UnityEngine.Pool;

namespace Magify
{
    internal class PurchaseTracker : ILegacyMigrator, IInitializable, IForegroundListener, IBackgroundListener, ICancelable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.Purchases);

        private readonly IServerApi _serverApi;
        private readonly PlatformAPI _platform;
        private readonly PurchasesStorage _storage;
        private readonly List<PurchaseRecord> _uploadingPurchaseRecords = new();
        private readonly SyncTimer _syncTimer;
        [NotNull]
        private readonly EmbeddedPurchaseVerificationHandler _embeddedPurchaseVerificationHandler = new();
        private List<PurchaseRecord> _unsentPurchaseRecords;
        private CompositeDisposable _disposables = GenericPool<CompositeDisposable>.Get();

        public bool IsDisposed => _disposables == null || _disposables.IsDisposed;

        [NotNull]
        public ReactiveProperty<IPurchaseVerificationHandler> ExternalPurchaseVerificationHandler { get; } = new(null);
        [NotNull]
        public FloatReactiveProperty VerificationRetryInterval { get; } = new(60.0f);
        private int RetryIntervalMills => Mathf.Max(Mathf.RoundToInt(VerificationRetryInterval.Value * 1000), 1000);

        public PurchaseTracker(IServerApi serverApi, PlatformAPI platform, string storagePath)
        {
            _serverApi = serverApi;
            _platform = platform;
            _storage = new PurchasesStorage(storagePath);
            _syncTimer = new SyncTimer(SendRecords).AddTo(_disposables);
            ExternalPurchaseVerificationHandler
                .Subscribe(handler => _logger.Log($"External purchase verification handler changed to {handler?.GetType().Name ?? "null"}"))
                .AddTo(_disposables);
            VerificationRetryInterval
                .Subscribe(_ =>
                {
                    _logger.Log($"Verification retry interval changed to {VerificationRetryInterval.Value} seconds");
                    _syncTimer?.Restart(RetryIntervalMills);
                })
                .AddTo(_disposables);
        }

        public void Initialize()
        {
            ThrowIfDisposed();
            _unsentPurchaseRecords = _storage.LoadRecords();
            _syncTimer.Start(RetryIntervalMills);
            SendRecords();
        }

        public void ValidateInApp(string productId, string transactionId = null, string purchaseToken = null, string receipt = null)
        {
            ValidatePurchase(productId, false, transactionId, purchaseToken, receipt);
        }

        public void ValidateSubscription(string productId, string transactionId = null, string purchaseToken = null, string receipt = null)
        {
            ValidatePurchase(productId, true, transactionId, purchaseToken, receipt);
        }

        public void ValidatePurchase(string productId, bool isSubscription, string transactionId = null, string purchaseToken = null, string receipt = null)
        {
            ThrowIfDisposed();
            var purchaseRecord = new PurchaseRecord
            {
                ProductId = productId,
#if UNITY_ANDROID
                IsSubscription = isSubscription,
                TransactionId = transactionId,
                PurchaseToken = purchaseToken,
#elif UNITY_IOS
                Receipt = receipt,
                DeviceId = _platform.GetEncodedDeviceIdentifier(),
                BuildNumber = _platform.BuildNumber,
#endif
            };
            AddUnsentPurchaseRecord(purchaseRecord);
            SendRecords();
            _syncTimer.Restart();
        }

        private void SendRecords()
        {
            for (var i = _unsentPurchaseRecords.Count - 1; i >= 0; i--)
            {
                var purchaseRecord = _unsentPurchaseRecords[i];
                SendRecord(purchaseRecord, _disposables.GetOrCreateToken()).Forget();
            }
        }

        private async UniTaskVoid SendRecord([NotNull] PurchaseRecord purchaseRecord, CancellationToken cancellationToken)
        {
            if (_uploadingPurchaseRecords.Contains(purchaseRecord)) return;

            var retryImmediately = false;
            try
            {
                _uploadingPurchaseRecords.Add(purchaseRecord);
                var result = await _serverApi.VerifyPurchase(purchaseRecord, waitForResult: ExternalPurchaseVerificationHandler.Value != null, cancellationToken);
                await TaskScheduler.SwitchToMainThread(cancellationToken);

                var handlerResult = ExternalPurchaseVerificationHandler.Value?.HandlePurchaseVerification(result)
                                            ?? _embeddedPurchaseVerificationHandler.HandlePurchaseVerification(result);
                _logger.Log($"For purchase {purchaseRecord.ProductId} received verification result: {result.Code}, handler result: {handlerResult}");
                if (handlerResult is RepeatState.Finish ||
                    result.Code is PurchaseVerificationResultCode.Success or PurchaseVerificationResultCode.Invalid)
                {
                    RemoveSentPurchaseRecord(purchaseRecord);
                }
                else if (handlerResult is RepeatState.Retry)
                {
                    retryImmediately = true;
                }
            }
            finally
            {
                if (!IsDisposed)
                {
                    _uploadingPurchaseRecords.Remove(purchaseRecord);
                    if (retryImmediately)
                    {
                        _logger.Log($"Immediately retrying verification of purchase {purchaseRecord.ProductId}");
                        SendRecord(purchaseRecord, cancellationToken).Forget();
                    }
                    else
                    {
                        _syncTimer.Restart();
                    }
                }
            }
        }

        private void AddUnsentPurchaseRecord(PurchaseRecord purchaseRecord)
        {
            _logger.Log($"Add unsent purchase record: {purchaseRecord}");
            _unsentPurchaseRecords.Add(purchaseRecord);
            _storage.SaveRecords(_unsentPurchaseRecords);
        }

        private void RemoveSentPurchaseRecord(PurchaseRecord purchaseRecord)
        {
            _logger.Log($"Remove sent purchase record: {purchaseRecord}");
            _unsentPurchaseRecords.Remove(purchaseRecord);
            _storage.SaveRecords(_unsentPurchaseRecords);
        }

        void IForegroundListener.OnForeground()
        {
            _syncTimer.Start(RetryIntervalMills);
        }

        void IBackgroundListener.OnBackground()
        {
            _syncTimer.Stop();
            _storage.SaveRecords(_unsentPurchaseRecords);
        }

        private void ThrowIfDisposed()
        {
            if (IsDisposed)
            {
                throw new ObjectDisposedException(nameof(PurchaseTracker));
            }
        }

        public void Dispose()
        {
            ThrowIfDisposed();
            _disposables.Clear();
            GenericPool<CompositeDisposable>.Release(_disposables);
            _disposables = default;
            _uploadingPurchaseRecords.Clear();
        }

        public void Migrate(MigrationData data)
        {
            _logger.Log("Try to migrate purchases from native");
            if (_storage.Exists())
            {
                _logger.Log("Saved purchases already exists. Skip migration.");
                return;
            }
            if (data.PurchasedRecords == null)
            {
                _logger.Log("No purchases to migration data. Skip migration.");
                return;
            }
            _storage.SaveRecords(data.PurchasedRecords!);
            _logger.Log($"Purchases successfully migrated in folder: {_storage.RootFolderPath}");
        }
    }
}