﻿using System;
using System.Linq;
using System.Threading;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using Magify.Model;
using NUnit.Framework;
using UnityEngine;
using UnityEngine.Networking;

namespace Magify.Tests
{
    internal partial class MinimalMagifyClientTests
    {
        internal class MinimalMagifyClientTestsPart
        {
            [SetUp]
            [TearDown]
            public void ClearAllData()
            {
                EditorModeTestsEnvironment.Clear();
            }

            protected CancellationToken GetCancellationToken(int cancelDelay = 1000)
            {
                return new CancellationTokenSource(cancelDelay).Token;
            }
        }

        private const string AppName = "TestApp";
        private const Environment OfflineEnvironment = Environment.Offline;

        [NotNull]
        private static MinimalMagifyClient CreateDefault(Environment? environment = Environment.Staging, INetworkClient networkClient = default, EditorDevice editorDevice = default)
        {
            editorDevice ??= EditorDevice.GetDefaultAndroidDevice();
            return networkClient == null
                ? new MinimalMagifyClient(AppName, EditorModeTestsEnvironment.RootStoragePath, environment ?? OfflineEnvironment)
                : new MinimalMagifyClient(AppName, EditorModeTestsEnvironment.RootStoragePath, environment ?? OfflineEnvironment, editorDevice, networkClient);
        }

        [NotNull]
        private static MinimalMagifyClient CreateDefaultAndInit(Environment? environment = Environment.Staging, INetworkClient networkClient = default, EditorDevice editorDevice = default)
        {
            editorDevice ??= EditorDevice.GetDefaultAndroidDevice();
            var client = networkClient == null
                ? new MinimalMagifyClient(AppName, EditorModeTestsEnvironment.RootStoragePath, environment ?? OfflineEnvironment)
                : new MinimalMagifyClient(AppName, EditorModeTestsEnvironment.RootStoragePath, environment ?? OfflineEnvironment, editorDevice, networkClient);
            client.InitializeSdk();
            return client;
        }

        [NotNull]
        private static MinimalAnalyticsTracker GetAnalyticsTracker([NotNull] MinimalMagifyClient magifyClient)
        {
            return GetPrivateField<MinimalAnalyticsTracker>(magifyClient, "_analyticsTracker");
        }

        [NotNull]
        private static ServerApi GetServerApi([NotNull] MinimalMagifyClient magifyClient)
        {
            return GetPrivateField<ServerApi>(magifyClient, "_serverApi");
        }

        [NotNull]
        private static MinimalAnalyticsService GetMinimalAnalyticsService([NotNull] MinimalMagifyClient magifyClient)
        {
            return GetPrivateField<MinimalAnalyticsService>(magifyClient, "_analyticsService");
        }

        [NotNull]
        private static PlatformAPI GetPlatformAPI([NotNull] MinimalMagifyClient magifyClient)
        {
            return GetPrivateField<PlatformAPI>(magifyClient, "_platform");
        }

        [NotNull]
        private static GeneralPrefs GetGeneralPrefs([NotNull] MinimalMagifyClient magifyClient)
        {
            return GetPrivateField<GeneralPrefs>(magifyClient, "_generalPrefs");
        }

        [NotNull]
        private static SessionCounter GetSessionCounter([NotNull] MinimalMagifyClient magifyClient)
        {
            return GetPrivateField<SessionCounter>(magifyClient, "_sessionCounter");
        }

        [NotNull]
        private static PurchaseTracker GetPurchaseTracker([NotNull] MinimalMagifyClient magifyClient)
        {
            return GetPrivateField<PurchaseTracker>(magifyClient, "_purchaseTracker");
        }

        [NotNull]
        private static T GetPrivateField<T>([NotNull] MinimalMagifyClient client, [NotNull] string fieldName)
            where T : class
        {
            return (client
                    .GetType()
                    .GetField(fieldName, System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Instance)
                    !.GetValue(client)
                    as T)!;
        }

        private class NetworkMoq : INetworkClient
        {
            internal delegate WebResponseMessage Handler(string url, WebRequestMessage message, CancellationToken cancellationToken);
            private readonly Handler _handler;

            public NetworkMoq(Handler handler)
            {
                _handler = handler;
            }

            public UniTask<WebResponseMessage> SendAsync(string url, WebRequestMessage message, CancellationToken cancellationToken)
            {
                try
                {
                    var result = _handler(url, message, cancellationToken);
                    return cancellationToken.IsCancellationRequested
                        ? UniTask.FromCanceled<WebResponseMessage>(cancellationToken)
                        : UniTask.FromResult(result);
                }
                catch (Exception e)
                {
                    return UniTask.FromException<WebResponseMessage>(e);
                }
            }

            public static WebResponseMessage OkResult(WebRequestMessage message)
            {
                object text = string.Empty;
                if (message.Method == ServerApi.RequestTokenMethods.GetToken)
                    text = new TokenResponse { Id = "id", Token = new string(Enumerable.Range(0, 100).Select(_ => 'a').ToArray()), };

                return new WebResponseMessage
                {
                    Result = UnityWebRequest.Result.Success,
                    ResponseCode = 200,
                    Error = string.Empty,
                    Text = JsonFacade.SerializeObject(text),
                    RequestMessage = message,
                };
            }

            public static WebResponseMessage ProtocolErrorResult(WebRequestMessage message, string text)
            {
                return new WebResponseMessage
                {
                    Result = UnityWebRequest.Result.ProtocolError,
                    ResponseCode = 403,
                    Error = string.Empty,
                    Text = text,
                    RequestMessage = message,
                };
            }

            public static WebResponseMessage ProtocolErrorResult(WebRequestMessage message, ErrorResponse text)
            {
                return new WebResponseMessage
                {
                    Result = UnityWebRequest.Result.ProtocolError,
                    ResponseCode = 403,
                    Error = string.Empty,
                    Text = JsonFacade.SerializeObject(text),
                    RequestMessage = message,
                };
            }

            public static bool TryExtractEvent<T>(WebRequestMessage message, Magify.Model.EventType type, out T @event)
                where T: IAnalyticsEvent
            {
                if (message.Payload is StorePayload storePayload && storePayload.Model == type)
                {
                    @event = (T)storePayload.Objects.FirstOrDefault();
                    return @event != null;
                }
                @event = default;
                return false;
            }

            public static bool TryExtractToken(WebRequestMessage message, out AuthTokenPayload authTokenPayload)
            {
                authTokenPayload = message.Payload as AuthTokenPayload;
                return authTokenPayload != null;
            }

            public static bool TryExtractVerification(WebRequestMessage message, out VerifyPurchasePayload verifyPurchasePayload, RuntimePlatform platform = RuntimePlatform.Android)
            {
                if (message.Payload is VerifyPurchasePayload payload && message.Method == ServerApi.MainNetworkRequestMethods.VerifyPurchase(platform))
                {
                    verifyPurchasePayload = payload;
                    return true;
                }
                verifyPurchasePayload = default;
                return false;
            }
        }
    }
}