using System;
using System.IO;
using Magify.Model;
using FluentAssertions;
using Magify.Features;
using NUnit.Framework;
using UniRx;
using UnityEngine;

namespace Magify.Tests
{
    [TestOf(typeof(FeaturesProvider))]
    public class FeaturesProviderTests
    {
        private const string TestConfig = "JsonModelParse/test_config";
        private const string KeyBool = "key_bool";
        private const string KeyNumber = "key_double";
        private const string KeyString = "key_string";

        private GeneralPrefs _prefs;
        private StoredAppFeaturesProvider _storedAppFeatures;

        [SetUp]
        public void Setup()
        {
            EditorModeTestsEnvironment.Clear();
            _prefs = GeneralPrefs.Create(EditorModeTestsEnvironment.GeneralPrefsPath, EditorModeTestsEnvironment.LocalGeneralPrefsPath, new PlatformDefault());
            _storedAppFeatures = new StoredAppFeaturesProvider(_prefs);
        }

        [TearDown]
        public void Cleanup()
        {
            ((IDisposable)_prefs).Dispose();
            ((IDisposable)_storedAppFeatures).Dispose();
            EditorModeTestsEnvironment.Clear();
        }

        [Test]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetBool))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetNumber))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetString))]
        public void GetFeatures_AndProviderDontHaveFeatures_ThenShouldReturnEmptyDictionary()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);

            // Act
            ((IInitializable)updater).Initialize();

            // Assert
            provider.TryGetBool(KeyBool, out _).Should().BeFalse();
            provider.TryGetNumber(KeyNumber, out _).Should().BeFalse();
            provider.TryGetString(KeyString, out _).Should().BeFalse();
        }

        [Test]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.UpdateContext))]
        public void SetupFeatures_ThenShouldInvokeOnUpdateCallback()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);
            ((IInitializable)updater).Initialize();
            var json = Resources.Load<TextAsset>(TestConfig).text;
            var context = CampaignsContext.Deserialize(json);

            // Act
            using var updateListener = Observable.FromEvent(
                    handler => provider.OnUpdate += handler,
                    handler => provider.OnUpdate -= handler)
                .Subscribe(_ => updateCallback());

            provider.UpdateContext(context, ContextKind.Downloaded);

            // Assert
            Assert.Fail();

            void updateCallback()
            {
                Assert.Pass();
            }
        }

        [Test]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.UpdateContext))]
        public void SetupDefaultFeatures_ThenShouldInvokeOnUpdateCallback()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);
            ((IInitializable)updater).Initialize();
            var json = Resources.Load<TextAsset>(TestConfig).text;
            var context = CampaignsContext.Deserialize(json);

            // Act
            using var updateListener = Observable.FromEvent(
                    handler => provider.OnUpdate += handler,
                    handler => provider.OnUpdate -= handler)
                .Subscribe(_ => updateCallback());

            provider.UpdateContext(context, ContextKind.Default);

            // Assert
            Assert.Fail();

            void updateCallback()
            {
                Assert.Pass();
            }
        }

        [Test]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.GetBool))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.GetNumber))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.GetString))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetBool))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetNumber))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetString))]
        public void GetFeatures_AndFeaturesWasSetup_ThenShouldReturnCorrectFeatures()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);
            ((IInitializable)updater).Initialize();
            var json = Resources.Load<TextAsset>(TestConfig).text;
            var context = CampaignsContext.Deserialize(json);
            provider.UpdateContext(context, ContextKind.Downloaded);

            // Act
            // Assert
            provider.GetBool(KeyBool).Should().BeFalse();
            provider.GetNumber(KeyNumber).Should().Be(1.2);
            provider.GetString(KeyString).Should().Be("TESTcurrent conf 123-aaaaa");
            provider.TryGetBool(KeyBool, out var boolResult).Should().BeTrue();
            provider.TryGetNumber(KeyNumber, out var numberResult).Should().BeTrue();
            provider.TryGetString(KeyString, out var stringResult).Should().BeTrue();
            boolResult.Should().Be(false);
            numberResult.Should().Be(1.2);
            stringResult.Should().Be("TESTcurrent conf 123-aaaaa");
        }

        [Test]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.GetBool))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.GetNumber))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.GetString))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetBool))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetNumber))]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.TryGetString))]
        public void GetDefaultFeatures_AndFeaturesWasSetup_ThenShouldReturnCorrectFeatures()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);
            ((IInitializable)updater).Initialize();
            var json = Resources.Load<TextAsset>(TestConfig).text;
            var context = CampaignsContext.Deserialize(json);
            provider.UpdateContext(context, ContextKind.Default);

            // Act
            // Assert
            provider.GetBool(KeyBool).Should().BeFalse();
            provider.GetNumber(KeyNumber).Should().Be(1.2);
            provider.GetString(KeyString).Should().Be("TESTcurrent conf 123-aaaaa");
            provider.TryGetBool(KeyBool, out var boolResult).Should().BeTrue();
            provider.TryGetNumber(KeyNumber, out var numberResult).Should().BeTrue();
            provider.TryGetString(KeyString, out var stringResult).Should().BeTrue();
            boolResult.Should().Be(false);
            numberResult.Should().Be(1.2);
            stringResult.Should().Be("TESTcurrent conf 123-aaaaa");
        }

        [Test]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.UpdateContext))]
        public void SetupFeatures_AndSameFeaturesWasSetup_ThenShouldNotInvokeOnUpdateCallback()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);
            ((IInitializable)updater).Initialize();
            var json = Resources.Load<TextAsset>(TestConfig).text;
            var context = CampaignsContext.Deserialize(json);
            provider.UpdateContext(context, ContextKind.Downloaded);

            // Act
            using var updateListener = Observable.FromEvent(
                    handler => provider.OnUpdate += handler,
                    handler => provider.OnUpdate -= handler)
                .Subscribe(_ => updateCallback());

            provider.UpdateContext(context, ContextKind.Downloaded);

            // Assert
            Assert.Pass();

            void updateCallback()
            {
                Assert.Fail();
            }
        }

        [Test]
        [UnitTestTarget(typeof(FeaturesProvider), nameof(FeaturesProvider.UpdateContext))]
        public void SetupDefaultFeatures_AndSameFeaturesWasSetup_ThenShouldNotInvokeOnUpdateCallback()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);
            ((IInitializable)updater).Initialize();
            var json = Resources.Load<TextAsset>(TestConfig).text;
            var context = CampaignsContext.Deserialize(json);
            provider.UpdateContext(context, ContextKind.Default);

            // Act
            using var updateListener = Observable.FromEvent(
                    handler => provider.OnUpdate += handler,
                    handler => provider.OnUpdate -= handler)
                .Subscribe(_ => updateCallback());

            provider.UpdateContext(context, ContextKind.Default);

            // Assert
            Assert.Pass();

            void updateCallback()
            {
                Assert.Fail();
            }
        }

        [Test]
        public void ChangeInAppStatus_AndSomeFeaturesWasRemovedForNewUserStatus_ThenShouldInvokeOnUpdateCallback()
        {
            // Arrange
            using var provider = new FeaturesProvider(_prefs);
            using var updater = new FeaturesUpdater(provider, _storedAppFeatures, _prefs);
            ((IInitializable)updater).Initialize();
            var json = Resources.Load<TextAsset>(TestConfig).text;
            var context = CampaignsContext.Deserialize(json);
            provider.UpdateContext(context, ContextKind.Downloaded);

            // Act
            using var updateListener = Observable.FromEvent(
                    handler => provider.OnUpdate += handler,
                    handler => provider.OnUpdate -= handler)
                .Subscribe(_ => updateCallback());

            _prefs.SubscriptionStatus.Value = SubscriptionStatus.Paid;

            // Assert
            Assert.Fail();

            void updateCallback()
            {
                Assert.Pass();
            }
        }
    }
}